@@ -17,7 +17,7 @@ use warnings;
 
 {
     package Doit;
-    our $VERSION = '0.025_56';
+    our $VERSION = '0.025_57';
     $VERSION =~ s{_}{};
 
     use constant IS_WIN => $^O eq 'MSWin32';
@@ -163,7 +163,7 @@ use warnings;
 {
     package Doit::Util;
     use Exporter 'import';
-    our @EXPORT; BEGIN { @EXPORT = qw(in_directory new_scope_cleanup copy_stat get_sudo_cmd) }
+    our @EXPORT; BEGIN { @EXPORT = qw(in_directory new_scope_cleanup copy_stat get_sudo_cmd is_in_path) }
     $INC{'Doit/Util.pm'} = __FILE__; # XXX hack
     use Doit::Log;
 
@@ -234,6 +234,54 @@ use warnings;
 	return ('sudo');
     }
 
+    sub is_in_path {
+	my($prog) = @_;
+
+	if (!defined &_file_name_is_absolute) {
+	    if (eval { require File::Spec; defined &File::Spec::file_name_is_absolute }) {
+		*_file_name_is_absolute = \&File::Spec::file_name_is_absolute;
+	    } else {
+		*_file_name_is_absolute = sub {
+		    my $file = shift;
+		    my $r;
+		    if ($^O eq 'MSWin32') {
+			$r = ($file =~ m;^([a-z]:(/|\\)|\\\\|//);i);
+		    } else {
+			$r = ($file =~ m|^/|);
+		    }
+		    $r;
+		};
+	    }
+	}
+
+	if (_file_name_is_absolute($prog)) {
+	    if ($^O eq 'MSWin32') {
+		return $prog       if (-f $prog && -x $prog);
+		return "$prog.bat" if (-f "$prog.bat" && -x "$prog.bat");
+		return "$prog.com" if (-f "$prog.com" && -x "$prog.com");
+		return "$prog.exe" if (-f "$prog.exe" && -x "$prog.exe");
+		return "$prog.cmd" if (-f "$prog.cmd" && -x "$prog.cmd");
+	    } else {
+		return $prog if -f $prog and -x $prog;
+	    }
+	}
+	require Config;
+	%Config::Config = %Config::Config if 0; # cease -w
+	my $sep = $Config::Config{'path_sep'} || ':';
+	foreach (split(/$sep/o, $ENV{PATH})) {
+	    if ($^O eq 'MSWin32') {
+		# maybe use $ENV{PATHEXT} like maybe_command in ExtUtils/MM_Win32.pm?
+		return "$_\\$prog"     if (-f "$_\\$prog" && -x "$_\\$prog");
+		return "$_\\$prog.bat" if (-f "$_\\$prog.bat" && -x "$_\\$prog.bat");
+		return "$_\\$prog.com" if (-f "$_\\$prog.com" && -x "$_\\$prog.com");
+		return "$_\\$prog.exe" if (-f "$_\\$prog.exe" && -x "$_\\$prog.exe");
+		return "$_\\$prog.cmd" if (-f "$_\\$prog.cmd" && -x "$_\\$prog.cmd");
+	    } else {
+		return "$_/$prog" if (-x "$_/$prog" && !-d "$_/$prog");
+	    }
+	}
+	undef;
+    }
 }
 
 {
@@ -684,17 +732,7 @@ use warnings;
 				     if ($quiet) {
 					 "copy $from $real_to";
 				     } else {
-					 if (eval { require IPC::Run; 1 }) {
-					     my $diff;
-					     if (eval { IPC::Run::run(['diff', '-u', $real_to, $from], '>', \$diff); 1 }) {
-						 "copy $from $real_to\ndiff:\n$diff";
-					     } else {
-						 "copy $from $real_to\n(diff not available" . (!$diff_error_shown++ ? ", error: $@" : "") . ")";
-					     }
-					 } else {
-					     my $diffref = _qx('diff', '-u', $real_to, $from);
-					     "copy $from $real_to\ndiff:\n$$diffref";
-					 }
+					 "copy $from $real_to\ndiff:\n" . _diff_files($real_to, $from);
 				     }
 				 }
 			     },
@@ -767,6 +805,24 @@ use warnings;
 	}
     }
 
+    sub _open2 {
+	my($instr, @args) = @_;
+	@args = Doit::Win32Util::win32_quote_list(@args) if Doit::IS_WIN;
+
+	require IPC::Open2;
+
+	my($chld_out, $chld_in);
+	my $pid = IPC::Open2::open2($chld_out, $chld_in, @args);
+	print $chld_in $instr;
+	close $chld_in;
+	local $/;
+	my $buf = <$chld_out>;
+	close $chld_out;
+	waitpid $pid, 0;
+
+	$buf;
+    }
+
     sub cmd_open2 {
 	my($self, @args) = @_;
 	my %options; if (@args && ref $args[0] eq 'HASH') { %options = %{ shift @args } }
@@ -775,19 +831,8 @@ use warnings;
 	my $instr = delete $options{instr}; $instr = '' if !defined $instr;
 	error "Unhandled options: " . join(" ", %options) if %options;
 
-	@args = Doit::Win32Util::win32_quote_list(@args) if Doit::IS_WIN;
-
-	require IPC::Open2;
-
 	my $code = sub {
-	    my($chld_out, $chld_in);
-	    my $pid = IPC::Open2::open2($chld_out, $chld_in, @args);
-	    print $chld_in $instr;
-	    close $chld_in;
-	    local $/;
-	    my $buf = <$chld_out>;
-	    close $chld_out;
-	    waitpid $pid, 0;
+	    my $buf = _open2($instr, @args);
 	    $? == 0
 		or _handle_dollar_questionmark($quiet||$info ? (prefix_msg => "open2 command '@args' failed: ") : ());
 	    $buf;
@@ -808,6 +853,46 @@ use warnings;
 	$self->cmd_open2(\%options, @args);
     }
 
+    sub _open3 {
+	my($instr, @args) = @_;
+	@args = Doit::Win32Util::win32_quote_list(@args) if Doit::IS_WIN;
+
+	require IO::Select;
+	require IPC::Open3;
+	require Symbol;
+
+	my($chld_out, $chld_in, $chld_err);
+	$chld_err = Symbol::gensym();
+	my $pid = IPC::Open3::open3((defined $instr ? $chld_in : undef), $chld_out, $chld_err, @args);
+	if (defined $instr) {
+	    print $chld_in $instr;
+	    close $chld_in;
+	}
+
+	my $sel = IO::Select->new;
+	$sel->add($chld_out);
+	$sel->add($chld_err);
+
+	my %buf = ($chld_out => '', $chld_err => '');
+	while(my @ready_fhs = $sel->can_read()) {
+	    for my $ready_fh (@ready_fhs) {
+		my $buf = '';
+		while (sysread $ready_fh, $buf, 1024, length $buf) { }
+		if ($buf eq '') { # eof
+		    $sel->remove($ready_fh);
+		    $ready_fh->close;
+		    last if $sel->count == 0;
+		} else {
+		    $buf{$ready_fh} .= $buf;
+		}
+	    }
+	}
+
+	waitpid $pid, 0;
+
+	($buf{$chld_out}, $buf{$chld_err});
+    }
+
     sub cmd_open3 {
 	my($self, @args) = @_;
 	my %options; if (@args && ref $args[0] eq 'HASH') { %options = %{ shift @args } }
@@ -818,44 +903,11 @@ use warnings;
 	my $statusref = delete $options{statusref};
 	error "Unhandled options: " . join(" ", %options) if %options;
 
-	@args = Doit::Win32Util::win32_quote_list(@args) if Doit::IS_WIN;
-
-	require IO::Select;
-	require IPC::Open3;
-	require Symbol;
-
 	my $code = sub {
-	    my($chld_out, $chld_in, $chld_err);
-	    $chld_err = Symbol::gensym();
-	    my $pid = IPC::Open3::open3((defined $instr ? $chld_in : undef), $chld_out, $chld_err, @args);
-	    if (defined $instr) {
-		print $chld_in $instr;
-		close $chld_in;
-	    }
-
-	    my $sel = IO::Select->new;
-	    $sel->add($chld_out);
-	    $sel->add($chld_err);
-
-	    my %buf = ($chld_out => '', $chld_err => '');
-	    while(my @ready_fhs = $sel->can_read()) {
-		for my $ready_fh (@ready_fhs) {
-		    my $buf = '';
-		    while (sysread $ready_fh, $buf, 1024, length $buf) { }
-		    if ($buf eq '') { # eof
-			$sel->remove($ready_fh);
-			$ready_fh->close;
-			last if $sel->count == 0;
-		    } else {
-			$buf{$ready_fh} .= $buf;
-		    }
-		}
-	    }
-
-	    waitpid $pid, 0;
+	    my($stdout, $stderr) = _open3($instr, @args);
 
 	    if ($errref) {
-		$$errref = $buf{$chld_err};
+		$$errref = $stderr;
 	    }
 
 	    if ($statusref) {
@@ -866,7 +918,7 @@ use warnings;
 		}
 	    }
 
-	    $buf{$chld_out};
+	    $stdout;
 	};
 
 	my @commands;
@@ -1172,6 +1224,15 @@ use warnings;
 	}
     }
 
+    sub cmd_which {
+	my($self, @args) = @_;
+	if (@args != 1) {
+	    error "Expecting exactly one argument: command";
+	}
+	my $path = Doit::Util::is_in_path($args[0]);
+	Doit::Commands->new({ rv => $path, code => sub {} });
+    }
+
     sub cmd_write_binary {
 	my($self, @args) = @_;
 	my %options; if (@args && ref $args[0] eq 'HASH') { %options = %{ shift @args } }
@@ -1235,35 +1296,7 @@ use warnings;
 					 }
 				     } else {
 					 if ($need_diff) {
-					     if (eval { require IPC::Run; 1 }) { # no temporary file required
-						 my $diff;
-						 if (eval { IPC::Run::run(['diff', '-u', $filename, '-'], '<', \$content, '>', \$diff); 1 }) {
-						     "Replace existing file $filename with diff:\n$diff";
-						 } else {
-						     "(diff not available" . (!$diff_error_shown++ ? ", error: $@" : "") . ")";
-						 }
-					     } else {
-						 my $diff;
-						 if (eval { require File::Temp; 1 }) {
-						     my($tempfh,$tempfile) = File::Temp::tempfile(UNLINK => 1);
-						     print $tempfh $content;
-						     if (close $tempfh) {
-							 my $diffref = _qx('diff', '-u', $filename, $tempfile);
-							 $diff = $$diffref;
-							 unlink $tempfile;
-							 if (length $diff) {
-							     $diff = "Replace existing file $filename with diff:\n$diff";
-							 } else {
-							     $diff = "(diff not available, probably no diff utility installed)";
-							 }
-						     } else {
-							 $diff = "(diff not available, error in tempfile creation ($!))";
-						     }
-						 } else {
-						     $diff = "(diff not available, neither IPC::Run nor File::Temp available)";
-						 }
-						 $diff;
-					     }
+					     "Replace existing file $filename with diff:\n" . _diff_files($filename, \$content);
 					 } else {
 					     "Create new file $filename with content:\n$content";
 					 }
@@ -1479,15 +1512,7 @@ use warnings;
 				     or error "Can't rename $tmpfile to $file: $!";
 			     },
 			     msg => do {
-				 my $diff;
-				 if (eval { require IPC::Run; 1 }) {
-				     if (!eval { IPC::Run::run(['diff', '-u', $file, $tmpfile], '>', \$diff); 1 }) {
-					 $diff = "(diff not available" . (!$diff_error_shown++ ? ", error: $@" : "") . ")";
-				     }
-				 } else {
-				     $diff = `diff -u '$file' '$tmpfile'`;
-				 }
-				 "Final changes as diff:\n$diff";
+				 "Final changes as diff:\n" . _diff_files($file, $tmpfile);
 			     },
 			     rv => $no_of_changes,
 			    };
@@ -1500,6 +1525,42 @@ use warnings;
 	}
     }
 
+    sub _diff_files {
+	my($file1, $file2) = @_;
+
+	my $stdin;
+	if (ref $file2) {
+	    $stdin = $$file2;
+	    $file2 = '-';
+	}
+
+	my($diff, $diff_stderr);
+	if (eval { require IPC::Run; 1 }) {
+	    IPC::Run::run(['diff', '-u', $file1, $file2], (defined $stdin ? ('<', \$stdin) : ()), '>', \$diff, '2>', \$diff_stderr);
+	} else {
+	    if ($^O eq 'MSWin32') { # list systems with unreliable IPC::Open3 here
+		my $tmp;
+		if ($file2 eq '-') {
+		    require File::Temp;
+		    $tmp = File::Temp->new;
+		    binmode($tmp); # XXX yes or no?
+		    $tmp->print($stdin);
+		    $tmp->close;
+		    $file2 = "$tmp";
+		}
+		my $diffref = _qx('diff', '-u', $file1, $file2);
+		$diff = $$diffref;
+		$diff_stderr = '';
+	    } else {
+		($diff, $diff_stderr) = eval { _open3($stdin, 'diff', '-u', $file1, $file2) };
+		if ($@) {
+		    $diff = "(diff not available" . (!$diff_error_shown++ ? ", error: $@" : "") . ")";
+		}
+	    }
+	}
+	"$diff$diff_stderr";
+    }
+
 }
 
 {
@@ -1644,6 +1705,7 @@ use warnings;
 		 qw(cond_run), # conditional run
 		 qw(touch), # like unix touch
 		 qw(ln_nsf), # like unix ln -nsf
+		 qw(which), # like unix which
 		 qw(create_file_if_nonexisting), # does the half of touch
 		 qw(write_binary), # like File::Slurper
 		 qw(change_file), # own invention
